@@ -11,7 +11,7 @@ from watchdog.observers import Observer
 from watchdog.events import FileSystemEventHandler
 from typing import List, Set, Optional
 from openhab_pythonrule_engine.item_registry import ItemRegistry
-from openhab_pythonrule_engine.trigger import TriggerRegistry, CronTrigger, ItemChangedTrigger, SystemEventTrigger, Trigger, Execution
+from openhab_pythonrule_engine.trigger import TriggerRegistry, CronTrigger, ItemChangedTrigger, RuleLoadedTrigger, Trigger, Execution
 from openhab_pythonrule_engine.eventbus_consumer import EventConsumer
 
 
@@ -42,6 +42,7 @@ class CronScheduler:
 
     def __init__(self):
         self.cron_trigger_by_module = {}
+        self.__is_running = True
         self.thread = Thread(target=self.__process, daemon=True)
 
     def add_job(self, cron_trigger: CronTrigger):
@@ -55,17 +56,24 @@ class CronScheduler:
             del self.cron_trigger_by_module[module]
 
     def __process(self):
-        while True:
+        logging.info("cron scheduler started")
+        while self.__is_running:
             for cron_triggers in list(self.cron_trigger_by_module.values()):
                 for cron_trigger in list(cron_triggers):
                     if pycron.is_now(cron_trigger.cron):
                         try:
-                            logging.debug("executing rule " + cron_trigger.name + " (cron 'Time cron " + cron_trigger.cron + "')")
+                            logging.debug("executing rule " + cron_trigger.name + " (triggerred by 'Time cron " + cron_trigger.cron + "')")
                             cron_trigger.invoke(ItemRegistry.instance())
                         except Exception as e:
                             logging.warning("Error occurred by executing rule " + cron_trigger.name, e)
-            sleep(60)
+            sleep(60)  # minimum 60 sec!
 
+    def start(self):
+        self.thread.start()
+
+    def stop(self):
+        self.__is_running = False
+        Thread.join(self.thread)
 
 
 class Rule:
@@ -125,6 +133,7 @@ class Rule:
 
     def add_trigger(self, trigger):
         self.__triggers.append(trigger)
+        self.__triggers = sorted(self.__triggers)
         trigger.add_listener(self.on_trigger_executed)
 
     def on_trigger_executed(self, trigger: Trigger):
@@ -134,14 +143,14 @@ class Rule:
             except Exception as e:
                 logging.warning("error occurred by calling listener", e)
 
-    def __str__(self):
-        text = self.module + ".py#" + self.name
-        for trigger in self.__triggers:
-            text += "\r\n  * " + trigger.expression
-            for execution in trigger.last_executions:
-                text += "\r\n      * " + str(execution)
+    def __hash__(self):
+        return hash(self.__str__())
 
-        return  text
+    def __eq__(self, other):
+        return self.__str__() == other.__str__()
+
+    def __str__(self):
+        return self.module + ".py#" + self.name + " (" + ", ".join(["'" + trigger.expression + "'" for trigger in self.__triggers]) + ")"
 
     def __repr__(self):
         return self.__str__()
@@ -167,8 +176,10 @@ class RuleEngine:
         logging.info("connecting " + openhab_uri)
         ItemRegistry.new_singleton(openhab_uri, user, pwd)
         self.__event_consumer = EventConsumer(openhab_uri, self)
-        self.cron_scheduler = CronScheduler()
-        self.trigger_registry = TriggerRegistry()
+        self.__event_consumer.start()
+        self.__cron_scheduler = CronScheduler()
+        self.__cron_scheduler.start()
+        self.__trigger_registry = TriggerRegistry()
 
     def start(self):
         logging.info("rules directory is " + self.__python_rule_directory)
@@ -184,16 +195,16 @@ class RuleEngine:
 
         self.__event_consumer.start()
 
-    def add_cron_trigger(self, trigger: CronTrigger):
-        self.trigger_registry.register(trigger)
-        self.cron_scheduler.add_job(trigger)
-
-    def add_item_changed_trigger(self, trigger: ItemChangedTrigger):
-        self.trigger_registry.register(trigger)
-
-    def add_system_event_trigger(self, trigger: SystemEventTrigger):
-        trigger.invoke(ItemRegistry.instance())
-        self.trigger_registry.register(trigger)
+    def add_trigger(self, trigger: Trigger):
+        if trigger.is_valid():
+            logging.info(" * " + trigger.name + "(...): register trigger '" + trigger.expression + "'")
+            self.__trigger_registry.register(trigger)
+            if isinstance(trigger, CronTrigger):
+                self.__cron_scheduler.add_job(trigger)
+            elif isinstance(trigger, RuleLoadedTrigger):
+                trigger.invoke(ItemRegistry.instance())
+        else:
+            logging.warning("Unsupported function spec " + trigger.module + "#" + trigger.name + " Ignoring it")
 
     def load_module(self, filename: str):
         if filename.endswith(".py"):
@@ -201,8 +212,8 @@ class RuleEngine:
             # reload?
             if modulename in sys.modules:
                 logging.info("reload '" + filename + "'")
-                self.cron_scheduler.remove_jobs(modulename)
-                self.trigger_registry.deregister(modulename)
+                self.__cron_scheduler.remove_jobs(modulename)
+                self.__trigger_registry.deregister(modulename)
                 importlib.reload(sys.modules[modulename])
             else:
                 logging.info("load '" + filename + "'")
@@ -212,8 +223,8 @@ class RuleEngine:
         modulename = self.__filename_to_modulename(filename)
         if modulename in sys.modules:
             logging.info("unload '" + filename + "'")
-            self.cron_scheduler.remove_jobs(modulename)
-            self.trigger_registry.deregister(modulename)
+            self.__cron_scheduler.remove_jobs(modulename)
+            self.__trigger_registry.deregister(modulename)
             del sys.modules[modulename]
 
     def __filename_to_modulename(self, filename):
@@ -221,18 +232,20 @@ class RuleEngine:
 
     def on_event(self, event):
         ItemRegistry.instance().on_event(event)
-        for item_changed_trigger in self.trigger_registry.get_triggers_by_type(ItemChangedTrigger):
+        triggers = self.__trigger_registry.get_triggers_by_type(ItemChangedTrigger)
+        matching_triggers = [trigger for trigger in triggers if trigger.matches(event)]
+        for item_changed_trigger in matching_triggers:
             item_changed_trigger.on_event(event)
 
     @property
     def rules(self) -> Set[Rule]:
         rules = set()
-        for module in self.trigger_registry.triggers_by_module.keys():
-            triggers = self.trigger_registry.triggers_by_module[module]
+        for module in self.__trigger_registry.triggers_by_module.keys():
+            triggers = self.__trigger_registry.triggers_by_module[module]
             for func in [trigger.func for trigger in triggers]:
                 rule = Rule(func)
-                rules.add(rule)
                 for trigger in triggers:
                     if trigger.name == rule.name:
                         rule.add_trigger(trigger)
+                rules.add(rule)
         return rules
\ No newline at end of file
